<?php

namespace Drupal\ui_patterns_field_formatters\Plugin\Derivative;

use Drupal\Component\Plugin\Derivative\DeriverBase;
use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldConfigInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\EntityContextDefinition;
use Drupal\Core\Plugin\Discovery\ContainerDeriverInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\TypedData\DataDefinitionInterface;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Drupal\ui_patterns\PropTypePluginManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides ui patterns source definitions for every field.
 *
 * @internal
 *   Plugin derivers are internal.
 */
abstract class FieldPropertiesSourceDeriverBase extends DeriverBase implements ContainerDeriverInterface {

  use StringTranslationTrait;
  use LoggerChannelTrait;

  /**
   * Constructs new FieldBlockDeriver.
   *
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *   The entity field manager.
   * @param \Drupal\Core\TypedData\TypedDataManagerInterface $typedDataManager
   *   The typed data manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\ui_patterns\PropTypePluginManager $propTypePluginManager
   *   The ui patterns prop type plugin manager.
   */
  public function __construct(
    protected EntityFieldManagerInterface $entityFieldManager,
    protected TypedDataManagerInterface $typedDataManager,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected PropTypePluginManager $propTypePluginManager,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, $base_plugin_id) {
    return new static(
      $container->get('entity_field.manager'),
      $container->get('typed_data_manager'),
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.ui_patterns_prop_type')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getDerivativeDefinitions($base_plugin_definition): array {
    // For each fieldable entity type, get the field map.
    foreach ($this->entityFieldManager->getFieldMap() as $entity_type_id => $entity_field_map) {
      $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
      $field_storage_definitions = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id);
      // Iterate over each field for that Entity Type.
      foreach ($entity_field_map as $field_name => $field_info) {
        if (!array_key_exists($field_name, $field_storage_definitions)) {
          continue;
        }
        // Prepare the plugin definition with options in common
        // for all the derivatives below.
        $derivative_for_this_field = $this->getBaseDerivativeDefinitionForField(
          $base_plugin_definition, $entity_type, $field_info['bundles'], $field_name);
        $this->getDerivativeDefinitionsForField($derivative_for_this_field, $field_storage_definitions[$field_name]);
      }
    }
    return $this->derivatives;
  }

  /**
   * Base plugin definition with common options for plugins derived for a field.
   *
   * @param array $base_plugin_definition
   *   The definition array of the base plugin.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type id.
   * @param array $bundles
   *   Bundles for the field.
   * @param string $field_name
   *   The field name.
   *
   * @return array
   *   Partially filled plugin definition.
   */
  protected function getBaseDerivativeDefinitionForField(
    array $base_plugin_definition,
    EntityTypeInterface $entity_type,
    array $bundles,
    string $field_name,
  ): array {
    $entity_type_id = $entity_type->id();
    $entity_context = EntityContextDefinition::fromEntityTypeId($entity_type_id)
      ->setRequired()
      ->setLabel($entity_type->getLabel());
    $field_name_context = new ContextDefinition('string');
    $field_name_context
      ->setRequired()
      ->setDefaultValue($field_name)
      ->addConstraint('AllowedValues', [$field_name]);
    $bundle_context = new ContextDefinition('string');
    $bundle_context
      ->setRequired()
      ->addConstraint('AllowedValues', array_merge(array_values($bundles), [""]));
    $derivative_for_this_field = array_merge($base_plugin_definition, [
      // The base id for the derivatives.
      'id' => implode(PluginBase::DERIVATIVE_SEPARATOR, [
        $entity_type_id,
        $field_name,
      ]),
      // Partially filled in.
      'ui_patterns_field_formatter' => [
        'entity_type' => $entity_type_id,
        'field_name' => $field_name,
      ],
      'context_definitions' => [
        'entity' => $entity_context,
        'field_name' => $field_name_context,
        'bundle' => $bundle_context,
      ],
    ]);
    // Add config_dependencies.
    if (!array_key_exists("config_dependencies", $derivative_for_this_field)) {
      $derivative_for_this_field["config_dependencies"] = [];
    }
    foreach ($bundles as $bundle) {
      $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
      if (empty($field_definitions[$field_name])) {
        continue;
      }
      $field_definition = $field_definitions[$field_name];
      // Add a dependency on the field if it is configurable.
      if ($field_definition instanceof FieldConfigInterface) {
        $derivative_for_this_field['config_dependencies'][$field_definition->getConfigDependencyKey()][] =
          $field_definition->getConfigDependencyName();
      }
    }
    return $derivative_for_this_field;
  }

  /**
   * Generate the deriver for properties field.
   *
   * @param array $derivative_for_this_field
   *   Base deriver for field.
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage_definition
   *   The field storage definition.
   */
  public function getDerivativeDefinitionsForField(array $derivative_for_this_field, FieldStorageDefinitionInterface $field_storage_definition) {
    // Generate one plugin for each field property,
    // compatible only for the corresponding bundles.
    foreach ($field_storage_definition->getPropertyDefinitions() as $property_id => $property_definition) {
      // Skip entity reference
      // Description could have been more precise,
      // but at the price of loading lot of stuff for nothing.
      // @todo check UI/UX needs for improvement of this.
      $description = t('Property "@property" of field "@field',
        [
          '@property' => $property_definition->getLabel(),
          '@field' => $field_storage_definition->getLabel(),
        ]);
      $derivative = array_merge(
        $derivative_for_this_field,
        [
          "label" => t("Field property : @property", ["@property" => $property_definition->getLabel()]),
          // @todo ,
          "description" => $description,
          "prop_types" => $this->getPropTypesDeriver($property_definition),
        ]
      );
      $derivative['id'] = implode(PluginBase::DERIVATIVE_SEPARATOR, [
        $derivative['id'],
        $property_id,
      ]);
      $derivative['ui_patterns_field_formatter'] = array_merge($derivative['ui_patterns_field_formatter'], [
        'property' => $property_id,
      ]);
      $this->derivatives[$derivative['id']] = $derivative;
    }
  }

  /**
   * Get compatible ui patterns prop types given a property definition.
   *
   * @param \Drupal\Core\TypedData\DataDefinitionInterface $property_definition
   *   The field property definition.
   *
   * @return array
   *   Array of ui patterns prop types.
   */
  abstract protected function getPropTypesDeriver(DataDefinitionInterface $property_definition): array;

}
