<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\Field\FieldFormatter;

use Drupal\Core\Field\FormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\EntityContext;
use Drupal\ui_patterns\Form\ComponentSettingsFormBuilderTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for ui_patterns field formatter plugin.
 */
abstract class ComponentFormatterBase extends FormatterBase {

  use ComponentSettingsFormBuilderTrait;

  /**
   * The component plugin manager.
   *
   * @var \Drupal\ui_patterns\ComponentPluginManager
   */
  protected $componentPluginManager;
  /**
   * The context repository.
   *
   * @var \Drupal\Core\Plugin\Context\ContextRepositoryInterface
   */
  protected $contextRepository;
  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;
  /**
   * The sample entity generator.
   *
   * @var \Drupal\ui_patterns\Entity\SampleEntityGeneratorInterface
   */
  protected $sampleEntityGenerator;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings']
    );
    $instance->componentPluginManager = $container->get('plugin.manager.sdc');
    $instance->contextRepository = $container->get('context.repository');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->sampleEntityGenerator = $container->get('ui_patterns.sample_entity_generator');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = [];
    // Get list of component by manager to insert label.
    $components = $this->componentPluginManager->getDefinitions();
    $options = [];
    $options['_empty'] = $this->t('Empty');
    foreach ($components as $component_id => $component) {
      $options[$component_id] = $component['name'];
    }
    $settings = $this->getSetting('ui_patterns');
    $summary["selected"] = $this->t('No component selected.')->render();
    if (!empty($settings['component_id'])) {
      $summary["selected"] = $this->t('Component ":component" selected.', [':component' => $options[$settings['component_id']]])->render();
    }
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return parent::defaultSettings() + self::getComponentFormDefault();
  }

  /**
   * {@inheritdoc}
   */
  public function getComponentSettings(): array {
    if (!empty($this->configuration)) {
      return $this->configuration;
    }
    return $this->getSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    return $this->componentSettingsForm($form, $form_state, $this->getComponentSourceContexts());
  }

  /**
   * Set the context of field and entity (override the method trait).
   *
   * @return array
   *   Source contexts.
   */
  protected function getComponentSourceContexts(): array {
    $field_definition = $this->fieldDefinition;
    $entity_type_id = $field_definition->getTargetEntityTypeId();

    $contextEntityTypeDefinition = ContextDefinition::create('string');
    $contextEntityType = new Context($contextEntityTypeDefinition, $entity_type_id);
    $context_source['entity_type_id'] = $contextEntityType;

    $contextFieldNameDefinition = ContextDefinition::create('string');
    $contextFieldName = new Context($contextFieldNameDefinition, $field_definition->getName());
    $context_source['field_name'] = $contextFieldName;

    $contextFieldNameDefinition = ContextDefinition::create('string');
    $contextFieldName = new Context($contextFieldNameDefinition, 'field_formatter');
    $context_source['plugin_type'] = $contextFieldName;

    $contextBundleDefinition = ContextDefinition::create('string');
    $bundle = $field_definition->getTargetBundle();
    if (NULL === $bundle) {
      // Generate a default bundle when it is missing,
      // this covers contexts like the display of a field in a view.
      $bundle = $entity_type_id;
      $bundle_entity_type = $this->entityTypeManager->getDefinition($entity_type_id)->getBundleEntityType();
      if (NULL !== $bundle_entity_type) {
        $bunle_list = $this->entityTypeManager->getStorage($bundle_entity_type)->loadMultiple();
        if (count($bunle_list) > 0) {
          $bundle = array_values($bunle_list)[0]->id();
        }
      }
    }
    $contextBundle = new Context($contextBundleDefinition, $bundle);
    $context_source['bundle'] = $contextBundle;

    $contexts = $this->contextRepository->getAvailableContexts();
    $runtime_contexts = $this->contextRepository->getRuntimeContexts(array_keys($contexts));

    foreach ($runtime_contexts as $context) {
      if ($context->getContextDefinition()
        ->getDataType() === 'entity:' . $entity_type_id) {
        $context_source['entity'] = $context;
      }
    }

    if (!array_key_exists("entity", $context_source)) {
      $context_source['entity'] = EntityContext::fromEntity($this->sampleEntityGenerator->get($entity_type_id, $bundle));
    }
    return $context_source;
  }

  /**
   * Alter the configuration of the mapping, like index and langcode.
   *
   * @param array $specific_config
   *   Array of specific configuration to insert.
   *
   * @return array
   *   The configuration array.
   */
  protected function setComponentSettingsExtra(array $specific_config): array {
    $configuration = $this->getComponentSettings();
    if (!array_key_exists('ui_patterns', $configuration)) {
      return $configuration;
    }
    if (!array_key_exists('props', $configuration['ui_patterns']) && !array_key_exists('slots', $configuration['ui_patterns'])) {
      return $configuration;
    }
    // Loop on each ui patterns props.
    $prop_ids = array_keys($configuration['ui_patterns']['props']);
    foreach ($prop_ids as $prop_id) {
      $configuration['ui_patterns']['props'][$prop_id]['source']['ui_patterns_field_formatter'] = $specific_config;
    }
    foreach ($configuration['ui_patterns']['slots'] as $slots_id => $slots) {
      if (empty($slots['sources']) || !is_array($slots['sources'])) {
        continue;
      }
      foreach (array_keys($slots['sources']) as $source_id) {
        $configuration['ui_patterns']['slots'][$slots_id]['sources'][$source_id]['source']['ui_patterns_field_formatter'] = $specific_config;
      }
    }
    $this->setComponentConfiguration($configuration['ui_patterns']);
    return $configuration['ui_patterns'];
  }

}
