<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source;

use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Field\FormatterInterface;
use Drupal\Core\Field\FormatterPluginManager;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Component field formatter, manage the form settings.
 */
abstract class FieldFormatterFormBase extends FieldFormatterSourceBase {

  use LoggerChannelTrait;

  use FieldFormatterFormTrait;

  /**
   * The formatter plugin manager.
   *
   * @var \Drupal\Core\Field\FormatterPluginManager
   */
  protected FormatterPluginManager $formatterPluginManager;

  /**
   * The field type plugin manager.
   *
   * @var \Drupal\Core\Field\FieldTypePluginManagerInterface
   */
  protected FieldTypePluginManagerInterface $fieldTypePluginManager;

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition,
  ) {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('context.repository')
    );
    $instance->entityFieldManager = $container->get('entity_field.manager');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->formatterPluginManager = $container->get('plugin.manager.field.formatter');
    $instance->fieldTypePluginManager = $container->get('plugin.manager.field.field_type');
    $instance->sampleEntityGenerator = $container->get('ui_patterns.sample_entity_generator');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $form = parent::settingsForm($form, $form_state);
    $this->buildFieldFormatterForm($form, $form_state);
    return $form;
  }

  /**
   * Test if we can add the formatter field formatter.
   *
   * @return bool
   *   Result of the test.
   */
  private function isApplicableFormatterFieldFormatter($plugin_definition) {
    $field_formatter_metadata = $this->getCustomPluginMetadata('field_formatter');
    if (!$field_formatter_metadata) {
      return FALSE;
    }
    // Only available for slots.
    if (!empty($plugin_definition['prop_types'][0]) && $plugin_definition['prop_types'][0] === 'slot') {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Callback to build field formatter form.
   *
   * @param array $form
   *   The source plugin settings form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   *
   * @return bool
   *   True if the form was generated.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function buildFieldFormatterForm(array &$form, FormStateInterface $form_state) {
    $plugin_definition = $this->getPluginDefinition();
    if (!$this->isApplicableFormatterFieldFormatter($plugin_definition)) {
      return FALSE;
    }
    $field_definition = $this->getFieldDefinition();
    if (!$field_definition instanceof FieldDefinitionInterface) {
      return FALSE;
    }
    $field_storage = $field_definition->getFieldStorageDefinition();
    $this->generateFieldFormatterForm($form, $form_state, $field_definition, $field_storage);
    return TRUE;
  }

  /**
   * Generate the form formatter field formatter.
   *
   * @param array $form
   *   The source plugin settings form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage
   *   Field storage.
   *
   * @return bool
   *   False if can't generate.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function generateFieldFormatterForm(array &$form, FormStateInterface $form_state, FieldDefinitionInterface $field_definition, FieldStorageDefinitionInterface $field_storage): bool {
    $formatter_options = $this->getAvailableFormatterOptions($field_storage, $field_definition);

    if (empty($formatter_options)) {
      return FALSE;
    }
    // @todo remove ui patterns formatters from the list of options ?
    // Get the formatter type from configuration.
    $type_path = [
      'settings',
      'type',
    ];
    $formatter_type = $this->getSettingsFromConfiguration($type_path);

    // Get the formatter settings from configuration.
    $settings = $this->getSettingsFieldFormatter($form_state, $field_storage, $formatter_options, $formatter_type);
    $form['type'] = [
      '#type' => 'select',
      '#title' => $this->t('Formatter'),
      '#options' => $formatter_options,
      '#default_value' => $formatter_type,
      // Note: We cannot use ::foo syntax, because the form is the entity form
      // display.
      '#ajax' => [
        'callback' => [__CLASS__, 'onFormatterTypeChange'],
        'wrapper' => 'field-formatter-settings-ajax',
        'method' => 'replace',
      ],
    ];

    $options = [
      'field_definition' => BaseFieldDefinition::createFromFieldStorageDefinition($field_storage),
      'configuration' => [
        'type' => $formatter_type,
        'settings' => $settings,
        'label' => '',
        'weight' => 0,
      ],
      'view_mode' => '_custom',
    ];

    // Get the formatter settings form.
    $settings_form = ['#value' => []];
    if ($formatter = $this->formatterPluginManager->getInstance($options)) {
      $settings_form = $formatter->settingsForm([], $form_state);
    }

    $form['settings'] = $settings_form;
    $form['settings']['#prefix'] = '<div id="field-formatter-settings-ajax">' . ($form['settings']['#prefix'] ?? '');
    $form['settings']['#suffix'] = ($form['settings']['#suffix'] ?? '') . '</div>';
    return TRUE;
  }

  /**
   * Get the formatter settings from configuration.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage
   *   Field storage.
   * @param array $formatter_options
   *   Array of formatters options.
   * @param string|null $formatter_type
   *   The formatter name.
   *
   * @return array|mixed|null
   *   The field formatter settings.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getSettingsFieldFormatter(FormStateInterface $form_state, FieldStorageDefinitionInterface $field_storage, array $formatter_options, ?string $formatter_type = '') {
    if (!empty($formatter_type)) {
      $settings_path = [
        'settings',
        'settings',
      ];
      $settings = $this->getSettingsFromConfiguration($settings_path);
    }
    // Get default formatter type.
    if (empty($formatter_type) || !isset($formatter_options[$formatter_type])) {
      $formatter_type = $this->fieldTypePluginManager->getDefinition($field_storage->getType())['default_formatter'] ?? key($formatter_options);
      $settings = $this->formatterPluginManager->getDefaultSettings($field_storage->getType());
    }
    // Reset settings if we change the formatter.
    $trigerring_element = $form_state->getTriggeringElement();
    if (!empty($trigerring_element) && $trigerring_element['#value'] === $formatter_type) {
      $settings = $this->formatterPluginManager->getDefaultSettings($formatter_type);
    }
    if (empty($settings) && !empty($formatter_type)) {
      $settings = $this->formatterPluginManager->getDefaultSettings($formatter_type);
    }
    return $settings;
  }

  /**
   * Create an instance of field formatter.
   *
   * @param string $formatter_id
   *   The formatter id.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition of field to apply formatter.
   *
   * @return \Drupal\Core\Field\FormatterBase
   *   The field formatter plugin.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function createInstanceFormatter(string $formatter_id, FieldDefinitionInterface $field_definition) {
    // @todo Ensure it is right to empty all values here, see:
    // https://api.drupal.org/api/drupal/core%21lib%21Drupal%21Core%21Field%21FormatterPluginManager.php/class/FormatterPluginManager/8.2.x
    $configuration = [
      'field_definition' => $field_definition,
      'settings' => [],
      'label' => '',
      'view_mode' => '',
      'third_party_settings' => [],
    ];
    return $this->formatterPluginManager->createInstance($formatter_id, $configuration);
  }

  /**
   * Get all available formatters by loading available ones and filtering out.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionInterface $field_storage_definition
   *   The field storage definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   Field definition.
   *
   * @return string[]
   *   The field formatter labels keys by plugin ID.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  protected function getAvailableFormatterOptions(FieldStorageDefinitionInterface $field_storage_definition, FieldDefinitionInterface $field_definition): array {
    $formatters = $this->formatterPluginManager->getOptions($field_storage_definition->getType());
    $formatter_instances = [];
    foreach ($formatters as $formater_id => $formatter) {
      $formatter_instances[$formater_id] = $this->createInstanceFormatter($formater_id, $field_definition);
    }

    $filtered_formatter_instances = $this->filterFormatter($formatter_instances, $field_definition);
    $options = array_map(
      static function (FormatterInterface $formatter) {
        return $formatter->getPluginDefinition()['label'];
      }, $filtered_formatter_instances);

    // Remove field_link itself.
    if (array_key_exists('field_link', $options)) {
      unset($options['field_link']);
    }

    return $options;
  }

}
