<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelTrait;

/**
 * Component field formatter, handle the data source.
 */
abstract class FieldFormatterSourceBase extends FieldPropertySourceBase {

  use LoggerChannelTrait;

  /**
   * Ajax callback for fields with AJAX callback to update form substructure.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return array
   *   The replaced form substructure.
   */
  public static function onFormatterTypeChange(array $form, FormStateInterface $form_state): array {
    $triggeringElement = $form_state->getTriggeringElement();
    // Dynamically return the dependent ajax for elements based on the
    // triggering element. This shouldn't be done statically because
    // settings forms may be different, e.g. for layout builder, core, ...
    if (!empty($triggeringElement['#array_parents'])) {
      $subformKeys = $triggeringElement['#array_parents'];
      // Remove the triggering element itself and add the 'settings' below key.
      array_pop($subformKeys);
      $subformKeys[] = 'settings';
      // Return the subform:
      return NestedArray::getValue($form, $subformKeys);
    }
    return [];
  }

  /**
   * Render field item(s) with the field formatter.
   *
   * @param \Drupal\Core\Field\FieldItemListInterface $items
   *   Items.
   * @param int|null $field_delta
   *   Field delta.
   *
   * @return array
   *   Render array
   */
  private function viewFieldItems(FieldItemListInterface $items, $field_delta = NULL): array {
    $returned = [];
    $configuration = $this->getConfiguration();
    if (empty($configuration['settings']['type'])) {
      // No formatter has been configured.
      return $returned;
    }
    for ($delta = 0; $delta < $items->count(); $delta++) {
      /** @var \Drupal\Core\Field\FieldItemInterface $item */
      $item = $items->get($delta);
      if ($field_delta !== NULL) {
        if ($delta !== $field_delta) {
          continue;
        }

        $returned[] = $item->view([
          'type' => $configuration['settings']['type'],
          'settings' => $configuration['settings']['settings'] ?? [],
        ]);
        break;
      }
      $returned[] = $item->view([
        'type' => $configuration['settings']['type'],
        'settings' => $configuration['settings']['settings'],
      ]);
    }
    return $returned;
  }

  /**
   * {@inheritdoc}
   */
  public function getPropValue(): mixed {
    $items = $this->getEntityFieldItemList();
    if (!$items instanceof FieldItemListInterface) {
      return [];
    }
    $field_index = $this->getConfiguredSettingsExtra('field_index') ?? NULL;
    return $this->viewFieldItems($items, $field_index);
  }

}
