<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_field_formatters\Plugin\UiPatterns\Source;

use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\ui_patterns\SourceInterface;
use Drupal\ui_patterns\SourcePluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Base class for all ui_patterns_field_formatter source plugins.
 */
abstract class FieldSourceBase extends SourcePluginBase implements SourceInterface {

  use LoggerChannelTrait;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition,
  ) {
    // We keep the same constructor as the parent class SourcePluginBase.
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('context.repository')
    );
    $instance->entityFieldManager = $container->get('entity_field.manager');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $form = parent::settingsForm($form, $form_state);
    $form['label_map'] = [
      "#type" => "label",
      "#title" => $this->propDefinition['title'] . ": " . $this->label(),
    ];
    return $form;
  }

  /**
   * Returns the field name.
   *
   * @return string
   *   The field name.
   */
  protected function getFieldName(): string {
    return $this->getContextValue('field_name');
  }

  /**
   * Returns the bundle.
   *
   * @return string
   *   The bundle.
   */
  protected function getBundle(): string {
    return $this->getContextValue('bundle');
  }

  /**
   * Returns the entity type id.
   *
   * @return string|null
   *   The entity type id.
   */
  protected function getEntityTypeId(): ?string {
    return $this->getContextValue('entity_type_id');
  }

  /**
   * Returns the field definition.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface|null
   *   The field definition.
   */
  protected function getFieldDefinition(): ?FieldDefinitionInterface {
    $entity_type_id = $this->getEntityTypeId();
    if (!$entity_type_id) {
      return NULL;
    }
    $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $this->getBundle());
    $field_name = $this->getFieldName();
    if (!array_key_exists($field_name, $field_definitions)) {
      return NULL;
    }
    return $field_definitions[$field_name];
  }

}
