<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_views\Plugin\UiPatterns\Source;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\EntityContextDefinition;
use Drupal\views\Plugin\views\field\FieldPluginBase;
use Drupal\views\ViewExecutable;

/**
 * Plugin base source to handle the view field.
 *
 * @Source(
 *   id = "view_field",
 *   label = @Translation("View field"),
 *    description = @Translation("View field."),
 *   prop_types = {
 *       "slot"
 *     },
 *  )
 */
class ViewFieldSource extends ViewsSourceBase {

  /**
   * {@inheritdoc}
   */
  public function getPropValue(): mixed {
    // Get field name inside the configuration.
    $parents = ['settings', 'ui_patterns_views_field'];
    $configuration = $this->getConfiguration();
    $field_name = NestedArray::getValue($configuration, $parents);
    // Get view data inside the configuration.
    $view = $this->getSettingsFromConfiguration('view');
    // Get row index inside the configuration.
    $row_index = $this->getSettingsFromConfiguration('row_index');
    if (empty($field_name) || empty($view) || (empty($row_index) && $row_index !== 0) || !$view instanceof ViewExecutable) {
      return ['#markup' => ''];
    }
    // Get the output of the field.
    $field_output = $view->style_plugin->getField($row_index, $field_name);
    // Remove field if need it.
    if ($this->removeEmptyField($field_name, $field_output, $view)) {
      return ['#markup' => ''];
    }
    return $this->renderOuput($field_output);
  }

  /**
   * Test if field must be deleted from render.
   *
   * @param string $field_name
   *   The field name.
   * @param mixed $field_output
   *   The render of field.
   * @param \Drupal\views\ViewExecutable $view
   *   The current view.
   *
   * @return bool
   *   Return TRUE if the field need to be removed from render.
   */
  private function removeEmptyField(string $field_name, mixed $field_output, ViewExecutable $view): bool {
    if (empty($field_name)) {
      return FALSE;
    }
    if (empty($view->field) || !is_array($view->field) || empty($view->field[$field_name]) || !($view->field[$field_name] instanceof FieldPluginBase)) {
      return TRUE;
    }
    $field = $view->field[$field_name];
    // Exclude field.
    if ($field->options['exclude']) {
      return TRUE;
    }
    $options = $this->getSettingsFromConfiguration('options');
    $empty_value = $field->isValueEmpty($field_output, $field->options['empty_zero']);
    // Remove field if empty.
    return $empty_value && ($field->options['hide_empty'] == TRUE || $options['hide_empty'] == TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function getContextDefinitions() {
    // Context for views entity type.
    $entity_context = EntityContextDefinition::fromEntityTypeId('view')
      ->setRequired()
      ->setLabel('Views');
    $plugin_type_context = new ContextDefinition('string');
    $plugin_type_context
      ->setRequired()
      ->addConstraint('AllowedValues', ['row']);
    return [
      'entity' => $entity_context,
      'plugin_type' => $plugin_type_context,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $form = parent::settingsForm($form, $form_state);
    $view = $this->getContextValue('entity');
    $current_display = $this->getContextValue('display');
    $default_settings = $this->getSettings();
    $options = [];
    foreach ($view->get('display') as $display_id => $display) {
      if ($current_display !== $display_id || empty($display['display_options']['fields'])) {
        continue;
      }

      foreach ($display['display_options']['fields'] as $field) {
        $options[$field['id']] = !empty($field['label']) ? $field['label'] : $field['field'];
      }
    }
    $form['ui_patterns_views_field'] = [
      '#type' => 'select',
      '#title' => t('Field'),
      '#description' => t('Select field to insert in this slot.'),
      '#options' => $options,
      '#default_value' => $default_settings['ui_patterns_views_field'] ?? [],
      '#required' => TRUE,
    ];
    return $form;
  }

}
