<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_views\Plugin\views\row;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\ui_patterns\Form\ComponentSettingsFormBuilderTrait;
use Drupal\views\Attribute\ViewsRow;
use Drupal\views\Plugin\views\row\Fields;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Pattern Views row plugin.
 *
 * @ingroup views_row_plugins
 */
#[ViewsRow(
  id: "ui_patterns",
  title: new TranslatableMarkup("Component (UI Patterns)"),
  help: new TranslatableMarkup("Displays fields using an UI component."),
  display_types: ['normal'],
  theme: "pattern_views_row",
  register_theme: FALSE,
)]
class ComponentRow extends Fields {

  use ComponentSettingsFormBuilderTrait;

  /**
   * The context repository.
   *
   * @var \Drupal\ui_patterns_views\UiPatternsViewsManager
   */
  protected $uiPatternsViewsManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
    );
    $instance->uiPatternsViewsManager = $container->get('ui_patterns_views.manager');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  protected function getAjaxUrl(FormStateInterface $form_state): ?Url {
    return $this->uiPatternsViewsManager->getViewsUiBuildFormUrl($form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function getComponentSettings(): array {
    if (!empty($this->configuration['ui_patterns'])) {
      return $this->configuration;
    }
    return $this->options;
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions(): array {
    return parent::defineOptions() + ['ui_patterns' => ['default' => self::getComponentFormDefault()['ui_patterns']]];
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);
    // Supported options.
    $keep_element_form = ['hide_empty'];
    $children = Element::children($form);
    foreach ($children as $child) {
      if (!in_array($child, $keep_element_form, TRUE)) {
        unset($form[$child]);
      }
    }
    // Build ui patterns component form.
    $ui_patterns_form = $this->componentSettingsForm($form, $form_state, $this->uiPatternsViewsManager->getComponentSourceContexts($this->view, $this));
    $form['ui_patterns'] = $ui_patterns_form['ui_patterns'];
  }

  /**
   * Remove the submit for inline row option.
   */
  public function submitOptionsForm(&$form, FormStateInterface $form_state) {}

  /**
   * {@inheritdoc}
   */
  public function render($row) {
    $configuration = $this->getComponentSettings();

    // Insert views data inside each prop configuration.
    foreach ($configuration['ui_patterns']['props'] as &$prop) {
      $prop['source']['ui_patterns_views'] = [
        'view' => $this->view,
        'row_index' => $row->index,
        'options' => $this->options,
      ];
    }
    unset($prop);
    // Insert views data inside each slot configuration.
    foreach ($configuration['ui_patterns']['slots'] as &$slots) {
      foreach ($slots['sources'] as &$source) {
        $source['source']['ui_patterns_views'] = [
          'view' => $this->view,
          'row_index' => $row->index,
          'options' => $this->options,
        ];
      }
    }
    unset($slots, $source);
    // Set the configuration.
    $this->setComponentConfiguration($configuration['ui_patterns']);
    // Build the component.
    $component = $this->buildComponentRenderable($this->getComponentConfiguration()['component_id'], $this->uiPatternsViewsManager->getComponentSourceContexts($this->view, $this));

    return [
      '#theme' => $this->themeFunctions(),
      '#view' => $this->view,
      '#options' => $this->options,
      '#row' => $row,
      '#pattern' => $component,
    ];
  }

}
