<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_views\Plugin\views\style;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\ui_patterns\Form\ComponentSettingsFormBuilderTrait;
use Drupal\views\Attribute\ViewsStyle;
use Drupal\views\Plugin\views\style\StylePluginBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Style plugin to render items in a pattern field.
 *
 * @ingroup views_style_plugins
 */
#[ViewsStyle(
  id: "ui_patterns",
  title: new TranslatableMarkup("Component (UI Patterns)"),
  help: new TranslatableMarkup("Displays views with UI components."),
  theme: "pattern_views_style",
  display_types: ["normal"],
  register_theme: FALSE,
)]
class ComponentStyle extends StylePluginBase {

  use ComponentSettingsFormBuilderTrait;

  /**
   * The context repository.
   *
   * @var \Drupal\ui_patterns_views\UiPatternsViewsManager
   */
  protected $uiPatternsViewsManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
    );
    $instance->uiPatternsViewsManager = $container->get('ui_patterns_views.manager');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  protected $usesRowPlugin = TRUE;

  /**
   * {@inheritdoc}
   */
  protected $usesOptions = TRUE;

  /**
   * {@inheritdoc}
   */
  protected function defineOptions(): array {
    return parent::defineOptions() + ['ui_patterns' => ['default' => self::getComponentFormDefault()['ui_patterns']]];
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);
    $this->uiPatternsViewsManager->getComponentSourceContexts($this->view, $this);

    $context_definition = $this->uiPatternsViewsManager->getComponentSourceContexts($this->view, $this);

    // Wrapper ajax id.
    $wrapper_id = '';
    $group_title = $this->addAjaxGroupTitleOption($wrapper_id, $form, $form_state);
    // Rebuild the current context depending on the group title.
    $context_group_title = ContextDefinition::create('boolean');
    $context_definition['group_title'] = new Context($context_group_title, !empty($group_title));

    // Build ui patterns component form.
    $ui_patterns_form = $this->componentSettingsForm($form, $form_state, $context_definition);
    $form['ui_patterns'] = $ui_patterns_form['ui_patterns'];
    // Add wrapper for the callback ajax (no need if group field option is not
    // present.
    if (!empty($wrapper_id)) {
      $form['ui_patterns']['#prefix'] = '<div id="' . $wrapper_id . '">';
      $form['ui_patterns']['#sufffix'] = '</div>';
    }
  }

  /**
   * Add ajax callback on the group title option.
   *
   * @param string $wrapper_id
   *   The ajax wrapper id.
   * @param array $form
   *   The current style view form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state object.
   *
   * @return string
   *   The value of group title option.
   */
  private function addAjaxGroupTitleOption(string &$wrapper_id, array &$form, FormStateInterface $form_state): string {
    // Add ajax callback to rebuild ui patterns form given the group title
    // option.
    if (empty($form['grouping']) || !is_array($form['grouping'])) {
      return '';
    }
    $wrapper_id = 'ui_patterns_views_group_title';
    $style_option = $form_state->getValue('style_options');
    $group_title = '';
    $trigerring_element = $form_state->getTriggeringElement();
    foreach ($form['grouping'] as $delta_group => &$group) {
      // Add ajax callback.
      $group['field']['#ajax'] = [
        'callback' => [static::class, 'groupTitleAjax'],
        'wrapper' => $wrapper_id,
        'effect' => 'fade',
      ];

      // Search for a group title option set.
      if (!empty($trigerring_element) && !empty($style_option['grouping'][$delta_group]['field'])) {
        $group_title = $style_option['grouping'][$delta_group]['field'];
      }
      if (empty($trigerring_element) && !empty($group['field']['#default_value'])) {
        $group_title = $group['field']['#default_value'];
      }
    }
    unset($group);
    // If group title found return it directly (no need to remove field from
    // the mapping.
    if (!empty($group_title)) {
      return $group_title;
    }
    // No group title, rebuild mapping (remove source group title).
    $this->removeGroupTitleSource($style_option['ui_patterns']['props']);
    foreach ($style_option['ui_patterns']['slots'] as &$slots) {
      foreach ($slots as &$sources) {
        $this->removeGroupTitleSource($sources);
      }
    }
    unset($slots, $sources);
    // Set value inside form state.
    $form_state->setValue('style_options', $style_option);
    $user_input = $form_state->getUserInput();
    $style_option_input = $user_input['style_options'] ?? [];
    // Set input inside form state.
    if (!empty($style_option_input)) {
      $this->removeGroupTitleSource($style_option_input['ui_patterns']['component_form']['props']);
      foreach ($style_option_input['ui_patterns']['component_form']['slots'] as &$slots) {
        foreach ($slots as &$sources) {
          $this->removeGroupTitleSource($sources);
        }
      }
      $user_input['style_options'] = $style_option_input;
      $form_state->setUserInput($user_input);
    }
    unset($slots, $sources);

    return $group_title;
  }

  /**
   * Remove from the source mapping group title plugin source.
   *
   * @param array $mapping_source
   *   The configuration of ui_patterns source.
   */
  private function removeGroupTitleSource(array &$mapping_source) {
    foreach ($mapping_source as $delta_source => $source) {
      if ($source['source_id'] === 'view_group_title') {
        unset($mapping_source[$delta_source]);
      }
    }
  }

  /**
   * Ajax callback for group title style option..
   */
  public static function groupTitleAjax(array $form, FormStateInterface $form_state) {
    return $form['options']['style_options']['ui_patterns'];
  }

  /**
   * {@inheritdoc}
   */
  public function getComponentSettings(): array {
    if (!empty($this->configuration['ui_patterns'])) {
      return $this->configuration;
    }
    return $this->options;
  }

  /**
   * {@inheritdoc}
   */
  protected function getAjaxUrl(FormStateInterface $form_state): ?Url {
    return $this->uiPatternsViewsManager->getViewsUiBuildFormUrl($form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function render(): array {
    $build = parent::render();

    // Insert views data inside each prop configuration.
    foreach ($build as $delta => $group) {
      $configuration = $this->getComponentSettings();
      foreach ($configuration['ui_patterns']['props'] as &$prop) {
        $prop['source']['ui_patterns_views'] = [
          'group' => $group,
          'view' => $this->view,
        ];
      }
      unset($prop);
      // Insert views data inside each slot configuration.
      foreach ($configuration['ui_patterns']['slots'] as &$slots) {
        foreach ($slots['sources'] as &$source) {
          $source['source']['ui_patterns_views'] = [
            'group' => $group,
            'view' => $this->view,
          ];
        }
      }
      unset($slots, $source);
      // Set the configuration.
      $this->setComponentConfiguration($configuration['ui_patterns']);
      // Build the component.
      $component = $this->buildComponentRenderable($this->getComponentConfiguration()['component_id'], $this->uiPatternsViewsManager->getComponentSourceContexts($this->view, $this));
      $build[$delta] = [
        '#theme' => $this->themeFunctions(),
        '#view' => $this->view,
        '#options' => $this->options,
        '#group' => $group,
        '#pattern' => $component,
      ];
    }
    return $build;
  }

}
