<?php

declare(strict_types=1);

namespace Drupal\ui_patterns_views;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\EntityContext;
use Drupal\Core\Plugin\PluginBase;
use Drupal\views\ViewExecutable;

/**
 * Helper manager of views plugin.
 */
class UiPatternsViewsManager {

  /**
   * Constructs the object.
   */
  public function __construct(protected EntityTypeManagerInterface $entityTypeManager, protected ModuleHandlerInterface $moduleHandler) {}

  /**
   * Set the context of field and entity (override the method trait).
   *
   * @param \Drupal\views\ViewExecutable $view
   *   The view executable object.
   * @param \Drupal\Core\Plugin\PluginBase $plugin
   *   The view plugin.
   *
   * @return array
   *   Source contexts.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getComponentSourceContexts(ViewExecutable $view, PluginBase $plugin): array {
    $plugin_definition = $plugin->getPluginDefinition();

    if (empty($plugin_definition['plugin_type']) || empty($plugin_definition['provider']) || $plugin_definition['provider'] !== 'ui_patterns_views') {
      return [];
    }

    $context_source = [];
    // Build views plugin type context.
    $context_plugin_type = ContextDefinition::create('string');
    $context_source['plugin_type'] = new Context($context_plugin_type, $plugin_definition['plugin_type']);
    // Build context views id.
    $context_id = ContextDefinition::create('string');
    $context_source['views_id'] = new Context($context_id, $view->id());
    $context_display = ContextDefinition::create('string');

    // Build views display context.
    $display = $view->getDisplay();
    $display_id = $display->getPluginId();
    // Display has no override of field.
    if (!isset($display->displa['display_options']['fields'])) {
      $display_id = 'default';
      if (!empty($display->default_display)) {
        /** @var \Drupal\views\Plugin\views\display\DefaultDisplay $default_display */
        $default_display = $display->default_display;
        $display_id = $default_display->getPluginId();
      }
    }
    $context_source['display'] = new Context($context_display, $display_id);
    // Build view entity context.
    $view_entity = $this->entityTypeManager->getStorage('view')
      ->load($view->id());

    if (!empty($view_entity)) {
      $context_source['entity'] = EntityContext::fromEntity($view_entity);
    }
    return $context_source;
  }

  /**
   * Return the views ui ajax form url.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current form state.
   *
   * @return \Drupal\Core\Url|null
   *   The views ajax url.
   */
  public function getViewsUiBuildFormUrl(FormStateInterface $form_state) {
    if ($this->moduleHandler->moduleExists('views_ui') && function_exists('views_ui_build_form_url')) {
      return views_ui_build_form_url($form_state);
    }
    return NULL;
  }

}
