<?php

declare(strict_types=1);

namespace Drupal\ui_patterns;

use Drupal\Component\Discovery\YamlDirectoryDiscovery;
use Drupal\Component\Plugin\CategorizingPluginManagerInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Theme\Component\ComponentValidator;
use Drupal\Core\Theme\Component\SchemaCompatibilityChecker;
use Drupal\Core\Theme\ComponentNegotiator;
use Drupal\Core\Theme\ComponentPluginManager as SdcPluginManager;
use Drupal\Core\Theme\ExtensionType;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\ui_patterns\SchemaManager\ReferencesResolver;

/**
 * UI Patterns extension of SDC component plugin manager.
 */
class ComponentPluginManager extends SdcPluginManager implements CategorizingPluginManagerInterface {

  /**
   * Cache key prefix to use in the cache backend.
   */
  const CACHE_KEY = 'ui_patterns';

  /**
   * {@inheritdoc}
   */
  public function __construct(
    ModuleHandlerInterface $module_handler,
    ThemeHandlerInterface $themeHandler,
    CacheBackendInterface $cacheBackend,
    ConfigFactoryInterface $configFactory,
    ThemeManagerInterface $themeManager,
    ComponentNegotiator $componentNegotiator,
    FileSystemInterface $fileSystem,
    SchemaCompatibilityChecker $compatibilityChecker,
    ComponentValidator $componentValidator,
    string $appRoot,
    protected PropTypePluginManager $propTypePluginManager,
    protected PropTypeAdapterPluginManager $prop_type_adapterPluginManager,
    protected ReferencesResolver $referencesSolver,
  ) {
    parent::__construct(
      $module_handler,
      $themeHandler,
      $cacheBackend,
      $configFactory,
      $themeManager,
      $componentNegotiator,
      $fileSystem,
      $compatibilityChecker,
      $componentValidator,
      $appRoot
    );
    $this->setCacheBackend($cacheBackend, self::CACHE_KEY);
  }

  /**
   * {@inheritdoc}
   */
  protected function alterDefinition(array $definition): array {
    // Overriding SDC alterDefinition method.
    $definition['extension_type'] = $this->moduleHandler->moduleExists($definition['provider'])
      ? ExtensionType::Module
      : ExtensionType::Theme;
    $metadata_path = $definition[YamlDirectoryDiscovery::FILE_KEY];
    $component_directory = $this->fileSystem->dirname($metadata_path);
    $definition['path'] = $component_directory;
    [, $machine_name] = explode(':', $definition['id']);
    $definition['machineName'] = $machine_name;
    $definition['library'] = $this->libraryFromDefinition($definition);
    $template = $this->findAsset(
      $component_directory,
      $definition['machineName'],
      'twig'
    );
    // This is why we are overriding ...
    if ($template) {
      $definition['template'] = basename($template);
    }
    $definition['documentation'] = 'No documentation found. Add a README.md in your component directory.';
    $documentation_path = sprintf('%s/README.md', $this->fileSystem->dirname($metadata_path));
    if (file_exists($documentation_path)) {
      $definition['documentation'] = file_get_contents($documentation_path);
    }

    // Adding custom logic.
    $fallback_prop_type_id = $this->propTypePluginManager->getFallbackPluginId("");
    $definition["name"] = $definition["name"] ?? $machine_name;
    $definition = $this->alterSlots($definition);
    $definition = $this->annotateProps($definition, $fallback_prop_type_id);
    $definition = $this->resolveTemplatePath($definition);

    return $definition;
  }

  /**
   * Alter slots.
   */
  protected function alterSlots(array $definition): array {
    if (!isset($definition['slots'])) {
      return $definition;
    }
    // Prevent slots without title from breaking.
    foreach ($definition['slots'] as $slot_id => $slot) {
      $definition['slots'][$slot_id]["title"] = $slot["title"] ?? $slot_id;
    }
    return $definition;
  }

  /**
   * Annotate each prop in a component definition.
   *
   * This is the main purpose of overriding SDC component plugin manager.
   * We add a 'ui_patterns' object in each prop schema of the definition.
   */
  protected function annotateProps(array $definition, string $fallback_prop_type_id): array {
    if (!isset($definition['props']['properties'])) {
      return $definition;
    }
    foreach ($definition['props']['properties'] as $prop_id => $prop) {
      $definition["title"] = $definition["title"] ?? $prop_id;
      $prop_type = $this->propTypePluginManager->guessFromSchema($prop);
      if ($prop_type->getPluginId() === $fallback_prop_type_id) {
        // Sometimes, a prop JSON schema is different enough to not be caught by
        // the compatibility checker, but close enough to address the same
        // sources as an existing prop type with only some small unidirectional
        // transformation of the data. So, we need an adapter plugin.
        $prop_type_adapter = $this->prop_type_adapterPluginManager->guessFromSchema($prop);
        if ($prop_type_adapter) {
          $prop_type_id = $prop_type_adapter->getPropTypeId();
          $prop_type = $this->propTypePluginManager->createInstance($prop_type_id);
          $prop['ui_patterns']['prop_type_adapter'] = $prop_type_adapter->getPluginId();
        }
      }

      if (isset($prop['$ref']) && str_starts_with($prop['$ref'], "ui-patterns://")) {
        // Resolve prop schema here, because:
        // - Drupal\Core\Theme\Component\ComponentValidator::getClassProps() is
        //   executed before schema references are resolved, so SDC believe
        //   a reference is a PHP namespace.
        // - It is not possible to propose a patch to SDC because
        //   SchemaStorage::resolveRefSchema() is not recursively resolving
        //   the schemas anyway.
        $prop = $this->referencesSolver->resolve($prop);
      }
      $prop['ui_patterns']['type_definition'] = $prop_type;
      $prop['ui_patterns']["summary"] = $prop_type->getSummary($prop);
      $definition['props']['properties'][$prop_id] = $prop;
    }
    $slot_prop_type = $this->propTypePluginManager->createInstance('slot', []);

    if (!empty($definition['slots'])) {
      foreach ($definition['slots'] as $slot_id => $slot) {
        $slot['ui_patterns']['type_definition'] = $slot_prop_type;
        $definition['slots'][$slot_id] = $slot;
      }
    }

    return $definition;
  }

  /**
   * Finds assets related to the provided metadata file.
   *
   * @param string $component_directory
   *   The component directory for the plugin.
   * @param string $machine_name
   *   The component's machine name.
   * @param string $file_extension
   *   The file extension to detect.
   * @param bool $make_relative
   *   TRUE to make the filename relative to the core folder.
   *
   * @return string|null
   *   Filenames, maybe relative to the core folder.
   */
  private function findAsset(string $component_directory, string $machine_name, string $file_extension, bool $make_relative = FALSE): ?string {
    $absolute_path = sprintf('%s%s%s.%s', $component_directory, DIRECTORY_SEPARATOR, $machine_name, $file_extension);
    if (!file_exists($absolute_path)) {
      return NULL;
    }
    return $make_relative
      ? $this->makePathRelativeToLibraryRoot($absolute_path)
      : $absolute_path;
  }

  /**
   * Takes a path and makes it relative to the library provider.
   *
   * Drupal will take a path relative to the library provider in order to put
   * CSS and JS in the HTML page. Core is the provider for all the
   * auto-generated libraries for the components. This means that in order to
   * add <root>/themes/custom/my_theme/components/my-component/my-component.css
   * in the page, we need to crawl back up from <root>/core first:
   * ../themes/custom/my_theme/components/my-component/my-component.css.
   *
   * @param string $path
   *   The path to the file.
   *
   * @return string
   *   The path relative to the library provider root.
   */
  private function makePathRelativeToLibraryRoot(string $path): string {
    $path_from_root = str_starts_with($path, $this->appRoot)
      ? substr($path, strlen($this->appRoot) + 1)
      : $path;
    // The library owner is in <root>/core, so we need to go one level up to
    // find the app root.
    return '../' . $path_from_root;
  }

  /**
   * Resolve template path.
   *
   * SDC is not allowing to have a custom template property because it is always
   * overriding its value. So we use _template instead.
   * See: https://www.drupal.org/project/drupal/issues/3390717
   */
  protected function resolveTemplatePath(array $definition): array {
    if (!isset($definition["_template"])) {
      return $definition;
    }
    $definition["template"] = $definition["_template"];
    unset($definition["_template"]);
    if (str_starts_with($definition["template"], "@")) {
      // @todo resolve namespace.
    }
    return $definition;
  }

  /**
   * {@inheritdoc}
   */
  public function getCategories() {
    // Fetch all categories from definitions and remove duplicates.
    $categories = array_unique(array_values(array_map(function ($definition) {
      return $definition['group'] ?? "Other";
    }, $this->getDefinitions())));
    natcasesort($categories);
    return array_values($categories);
  }

  /**
   * {@inheritdoc}
   */
  public function getSortedDefinitions(array $definitions = NULL) {
    // Sort the plugins first by group, then by label.
    $definitions = $definitions ?? $this->getDefinitions();
    $label_key = 'name';
    uasort($definitions, function ($a, $b) use ($label_key) {
      $a_group = isset($a['group']) ? (string) $a['group'] : '';
      $b_group = isset($b['group']) ? (string) $b['group'] : '';
      if ($a_group !== $b_group) {
        return strnatcasecmp($a_group, $b_group);
      }
      return strnatcasecmp($a[$label_key], $b[$label_key]);
    });
    return $definitions;
  }

  /**
   * {@inheritdoc}
   */
  public function getGroupedDefinitions(?array $definitions = NULL): array {
    $definitions = $definitions ?: $this->getDefinitions();
    $groups = [];
    foreach ($definitions as $id => $definition) {
      $group = $definition["group"] ?? "Other";
      $groups[$group][$id] = $definition;
    }
    return $groups;
  }

}
